==> v10.5.3.zip: brewman/brewman/brewman/routers/reports/profit_loss.py <==
from datetime import date, datetime
from decimal import Decimal
from typing import List, Optional, Tuple

import brewman.schemas.profit_loss as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import desc, func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.account_type import AccountType
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...routers.reports.closing_stock import get_closing_stock, get_opening_stock
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.ProfitLoss)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["profit-&-loss"]),
) -> schemas.ProfitLoss:
    return schemas.ProfitLoss(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        body=[],
        footer=None,
    )


@router.get("/{start}/{finish}", response_model=schemas.ProfitLoss)
def report_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["profit-&-loss"]),
) -> schemas.ProfitLoss:
    with SessionFuture() as db:
        body, footer = build_profit_loss(
            datetime.strptime(start, "%d-%b-%Y"), datetime.strptime(finish, "%d-%b-%Y"), db
        )
        set_period(start, finish, request.session)
        return schemas.ProfitLoss(
            startDate=start,
            finishDate=finish,
            body=body,
            footer=footer,
        )


def build_profit_loss(
    start_date: date, finish_date: date, db: Session
) -> Tuple[List[schemas.ProfitLossItem], schemas.ProfitLossItem]:
    profit_type_list = (
        db.execute(select(AccountType.id).where(AccountType.balance_sheet == False)).scalars().all()  # noqa: E712
    )
    report: List[schemas.ProfitLossItem] = []
    groups = {}

    amount_sum = func.sum(Journal.amount * Journal.debit)
    query = db.execute(
        select(AccountBase, amount_sum)
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            AccountBase.type_id.in_(profit_type_list),
        )
        .group_by(AccountBase)
        .order_by(AccountBase.type_id, desc(func.abs(amount_sum)))
    ).all()

    # Get opening / closing stock
    opening_stock = get_opening_stock(start_date, db)
    closing_stock = get_closing_stock(finish_date, db)
    total_amount = (opening_stock - closing_stock) * -1

    report.append(schemas.ProfitLossItem(name="Opening Stock", amount=opening_stock * -1, order=200001))
    report.append(schemas.ProfitLossItem(name="Closing Stock", amount=closing_stock, order=290000))

    purchase_group = db.execute(select(AccountType).where(AccountType.name == "Purchase")).scalar_one()
    groups[purchase_group.id] = schemas.ProfitLossItem(
        group=purchase_group.name,
        total=total_amount,
        order=purchase_group.order,
    )

    counter = 0
    for account, amount in query:
        # Add Items
        amount *= -1
        total_amount += amount

        if amount != 0:
            counter += 10
            report.append(
                schemas.ProfitLossItem(
                    name=account.name,
                    amount=amount,
                    order=account.type_.order + counter,
                )
            )
            if account.type_.id in groups:
                groups[account.type_.id].total += amount
            else:
                groups[account.type_.id] = schemas.ProfitLossItem(
                    group=account.type_.name,
                    total=amount,
                    order=account.type_.order,
                )

    # Add Subtotals
    for item in groups.values():
        report.append(item)

    # Add Net
    footer = schemas.ProfitLossItem(
        group="Net Profit" if total_amount > 0 else "Net Loss",
        total=total_amount,
        order=1000000,
    )

    return sorted(report, key=lambda d: d.order), footer


def get_accumulated_profit(finish_date: date, db: Session) -> Decimal:
    type_list = (
        db.execute(select(AccountType.id).where(AccountType.balance_sheet == False)).scalars().all()  # noqa: E712
    )

    accumulated_profit: Optional[Decimal] = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            AccountBase.type_id.in_(type_list),
        )
    ).scalar()
    return Decimal(0) if accumulated_profit is None else accumulated_profit

==> v11.1.5.zip: brewman/brewman/brewman/routers/reports/profit_loss.py <==
from datetime import date, datetime
from decimal import Decimal

import brewman.schemas.profit_loss as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import desc, func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.account_type import AccountType
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...routers.reports.closing_stock import get_closing_stock, get_opening_stock
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.ProfitLoss)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["profit-&-loss"]),
) -> schemas.ProfitLoss:
    return schemas.ProfitLoss(
        start_date=get_start_date(request.session),
        finish_date=get_finish_date(request.session),
        body=[],
        footer=schemas.ProfitLossItem(group="", total=Decimal(0), order=0),
    )


@router.get("/{start}/{finish}", response_model=schemas.ProfitLoss)
def report_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["profit-&-loss"]),
) -> schemas.ProfitLoss:
    with SessionFuture() as db:
        body, footer = build_profit_loss(
            datetime.strptime(start, "%d-%b-%Y"), datetime.strptime(finish, "%d-%b-%Y"), db
        )
        set_period(start, finish, request.session)
        return schemas.ProfitLoss(
            start_date=start,
            finish_date=finish,
            body=body,
            footer=footer,
        )


def build_profit_loss(
    start_date: date, finish_date: date, db: Session
) -> tuple[list[schemas.ProfitLossItem], schemas.ProfitLossItem]:
    profit_type_list = (
        db.execute(select(AccountType.id).where(AccountType.balance_sheet == False)).scalars().all()  # noqa: E712
    )
    report: list[schemas.ProfitLossItem] = []
    groups = {}

    amount_sum = func.sum(Journal.amount * Journal.debit)
    query = db.execute(
        select(AccountBase, amount_sum)
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date_ >= start_date,
            Voucher.date_ <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            AccountBase.type_id.in_(profit_type_list),
        )
        .group_by(AccountBase)  # type: ignore
        .order_by(AccountBase.type_id, desc(func.abs(amount_sum)))
    ).all()

    # Get opening / closing stock
    opening_stock = get_opening_stock(start_date, db)
    closing_stock = get_closing_stock(finish_date, db)
    total_amount = (opening_stock - closing_stock) * -1

    report.append(schemas.ProfitLossItem(name="Opening Stock", amount=opening_stock * -1, order=200001))
    report.append(schemas.ProfitLossItem(name="Closing Stock", amount=closing_stock, order=290000))

    purchase_group = db.execute(select(AccountType).where(AccountType.name == "Purchase")).scalar_one()
    groups[purchase_group.id] = schemas.ProfitLossItem(
        group=purchase_group.name,
        total=total_amount,
        order=purchase_group.order,
    )

    counter = 0
    for account, amount in query:
        # Add Items
        amount *= -1
        total_amount += amount

        if amount != 0:
            counter += 10
            report.append(
                schemas.ProfitLossItem(
                    name=account.name,
                    amount=amount,
                    order=account.type_.order + counter,
                )
            )
            if account.type_.id in groups:
                groups[account.type_.id].total += amount
            else:
                groups[account.type_.id] = schemas.ProfitLossItem(
                    group=account.type_.name,
                    total=amount,
                    order=account.type_.order,
                )

    # Add Subtotals
    for item in groups.values():
        report.append(item)

    # Add Net
    footer = schemas.ProfitLossItem(
        group="Net Profit" if total_amount > 0 else "Net Loss",
        total=total_amount,
        order=1000000,
    )

    return sorted(report, key=lambda d: d.order), footer


def get_accumulated_profit(finish_date: date, db: Session) -> Decimal:
    type_list = (
        db.execute(select(AccountType.id).where(AccountType.balance_sheet == False)).scalars().all()  # noqa: E712
    )

    accumulated_profit: Decimal | None = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date_ <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            AccountBase.type_id.in_(type_list),
        )
    ).scalar()
    return Decimal(0) if accumulated_profit is None else accumulated_profit
